"""
Comprehensive tests for all exception types in primp.

Tests each exception type with real scenarios that trigger them,
using the local test server to raise real exceptions.
"""
import pytest

import primp
from primp import (
    PrimpError,
    BuilderError,
    RequestError,
    ConnectError,
    TimeoutError,
    StatusError,
    RedirectError,
    BodyError,
    DecodeError,
    UpgradeError,
)


class TestExceptionHierarchy:
    """Test that the exception hierarchy is correct."""

    def test_primp_error_is_base(self):
        """PrimpError should be the base exception for all primp exceptions."""
        assert issubclass(BuilderError, PrimpError)
        assert issubclass(RequestError, PrimpError)
        assert issubclass(StatusError, PrimpError)
        assert issubclass(RedirectError, PrimpError)
        assert issubclass(BodyError, PrimpError)
        assert issubclass(DecodeError, PrimpError)
        assert issubclass(UpgradeError, PrimpError)

    def test_request_error_children(self):
        """RequestError should have ConnectError and TimeoutError as children."""
        assert issubclass(ConnectError, RequestError)
        assert issubclass(TimeoutError, RequestError)


class TestBuilderError:
    """Test BuilderError exception with real scenarios.

    BuilderError is raised for URL and header errors, providing a descriptive
    message explaining the specific error.
    """

    def test_missing_scheme(self):
        """URL without scheme should raise BuilderError."""
        client = primp.Client()
        with pytest.raises(BuilderError):
            client.get("example.com")

    def test_empty_url(self):
        """Empty URL should raise BuilderError."""
        client = primp.Client()
        with pytest.raises(BuilderError):
            client.get("")

    def test_invalid_url_format(self):
        """Invalid URL format should raise BuilderError."""
        client = primp.Client()
        with pytest.raises(BuilderError):
            client.get("://invalid")

    def test_can_catch_as_primp_error(self):
        """BuilderError should be catchable as PrimpError."""
        client = primp.Client()
        with pytest.raises(PrimpError):
            client.get("example.com")

    def test_invalid_header_with_newline(self, test_server):
        """Header with newline character should raise BuilderError."""
        client = primp.Client()
        # Header values with newlines are invalid
        with pytest.raises(BuilderError):
            client.get(f"{test_server}/get", headers={"X-Test": "value\nwith-newline"})

    def test_invalid_header_with_carriage_return(self, test_server):
        """Header with carriage return should raise BuilderError."""
        client = primp.Client()
        with pytest.raises(BuilderError):
            client.get(f"{test_server}/get", headers={"X-Test": "value\rwith-cr"})

    def test_invalid_header_name(self, test_server):
        """Invalid header name should raise BuilderError."""
        client = primp.Client()
        # Header names with special characters are invalid
        with pytest.raises(BuilderError):
            client.get(f"{test_server}/get", headers={"Invalid:Name": "value"})


class TestConnectError:
    """Test ConnectError exception with real scenarios."""

    def test_connection_refused(self):
        """Connection to non-listening port should raise ConnectError."""
        client = primp.Client()
        with pytest.raises(ConnectError):
            client.get("http://localhost:1", timeout=5)

    def test_can_catch_as_request_error(self):
        """ConnectError should be catchable as RequestError."""
        client = primp.Client()
        with pytest.raises(RequestError):
            client.get("http://localhost:1", timeout=0.5)

    def test_can_catch_as_primp_error(self):
        """ConnectError should be catchable as PrimpError."""
        client = primp.Client()
        with pytest.raises(PrimpError):
            client.get("http://localhost:1", timeout=0.5)


class TestTimeoutError:
    """Test TimeoutError exception with real scenarios."""

    def test_request_timeout(self, test_server):
        """Request that exceeds timeout should raise TimeoutError."""
        client = primp.Client()
        # Request a 5 second delay with a 0.5 second timeout
        with pytest.raises(TimeoutError):
            client.get(f"{test_server}/delay/5", timeout=0.5)

    def test_can_catch_as_request_error(self, test_server):
        """TimeoutError should be catchable as RequestError."""
        client = primp.Client()
        with pytest.raises(RequestError):
            client.get(f"{test_server}/delay/5", timeout=0.5)

    def test_can_catch_as_primp_error(self, test_server):
        """TimeoutError should be catchable as PrimpError."""
        client = primp.Client()
        with pytest.raises(PrimpError):
            client.get(f"{test_server}/delay/5", timeout=0.5)


class TestStatusError:
    """Test StatusError exception with real scenarios."""

    def test_404_status(self, test_server):
        """HTTP 404 should raise StatusError when raise_for_status is called."""
        client = primp.Client()
        response = client.get(f"{test_server}/status/404")
        with pytest.raises(StatusError):
            response.raise_for_status()

    def test_500_status(self, test_server):
        """HTTP 500 should raise StatusError when raise_for_status is called."""
        client = primp.Client()
        response = client.get(f"{test_server}/status/500")
        with pytest.raises(StatusError):
            response.raise_for_status()

    def test_403_status(self, test_server):
        """HTTP 403 should raise StatusError when raise_for_status is called."""
        client = primp.Client()
        response = client.get(f"{test_server}/status/403")
        with pytest.raises(StatusError):
            response.raise_for_status()

    def test_status_error_has_status_code(self, test_server):
        """StatusError args should contain status code."""
        client = primp.Client()
        response = client.get(f"{test_server}/status/404")
        with pytest.raises(StatusError) as exc_info:
            response.raise_for_status()
        # StatusError args: (status_code, message, url)
        assert exc_info.value.args[0] == 404

    def test_status_error_status_code_500(self, test_server):
        """StatusError args should contain status code for 500."""
        client = primp.Client()
        response = client.get(f"{test_server}/status/500")
        with pytest.raises(StatusError) as exc_info:
            response.raise_for_status()
        assert exc_info.value.args[0] == 500

    def test_can_catch_as_primp_error(self, test_server):
        """StatusError should be catchable as PrimpError."""
        client = primp.Client()
        response = client.get(f"{test_server}/status/404")
        with pytest.raises(PrimpError):
            response.raise_for_status()

    def test_successful_status_no_error(self, test_server):
        """Successful status codes should not raise StatusError."""
        client = primp.Client()
        response = client.get(f"{test_server}/status/200")
        # Should not raise
        response.raise_for_status()


class TestRedirectError:
    """Test RedirectError exception with real scenarios."""

    def test_max_redirects_exceeded(self, test_server):
        """Exceeding max_redirects should raise RedirectError."""
        client = primp.Client(follow_redirects=True, max_redirects=2)
        # Request 10 redirects with max_redirects=2
        with pytest.raises(RedirectError):
            client.get(f"{test_server}/redirect/10")

    def test_can_catch_as_primp_error(self, test_server):
        """RedirectError should be catchable as PrimpError."""
        client = primp.Client(follow_redirects=True, max_redirects=2)
        with pytest.raises(PrimpError):
            client.get(f"{test_server}/redirect/10")


class TestDecodeError:
    """Test DecodeError exception with real scenarios.

    DecodeError is raised when the server returns content with a
    Content-Encoding header (like gzip or deflate) but the actual
    content is not properly encoded. The client automatically decompresses
    based on the Content-Encoding header.
    """

    def test_invalid_gzip_content(self, test_server):
        """Invalid gzip content should raise DecodeError when decompressing."""
        client = primp.Client()
        # The server returns invalid gzip data with Content-Encoding: gzip
        # This should trigger DecodeError when the client tries to decompress
        # The error is raised when accessing the content, not during the request
        response = client.get(f"{test_server}/invalid-gzip")
        with pytest.raises(DecodeError):
            _ = response.content

    def test_invalid_deflate_content(self, test_server):
        """Invalid deflate content should raise DecodeError when decompressing."""
        client = primp.Client()
        # The server returns invalid deflate data with Content-Encoding: deflate
        response = client.get(f"{test_server}/invalid-deflate")
        with pytest.raises(DecodeError):
            _ = response.content

    def test_can_catch_as_primp_error(self, test_server):
        """DecodeError should be catchable as PrimpError."""
        client = primp.Client()
        response = client.get(f"{test_server}/invalid-gzip")
        with pytest.raises(PrimpError):
            _ = response.content


class TestBodyError:
    """Test BodyError exception.

    BodyError is raised when there's an error with the request or response body,
    such as when streaming a request body that encounters an error, or when
    there's a connection error while reading/writing the body.

    BodyError can be triggered by:
    1. Timeout while reading/writing body
    2. Connection errors while streaming body
    3. Request body stream errors

    Note: DecodeError is raised for transfer encoding errors (like broken chunked encoding),
    not BodyError. BodyError is for lower-level body I/O issues.
    """

    def test_request_body_error_on_closed_connection(self, test_server):
        """Sending a large request body to a server that closes early may trigger BodyError.

        This test attempts to trigger BodyError by sending a large request body to a server
        that closes the connection immediately. Depending on timing, this may result in
        BodyError, RequestError, or no error if the server responds before the body is sent.
        """
        client = primp.Client(timeout=2)
        # Create a large body that will take time to send
        large_body = b"x" * 10_000_000  # 10MB of data
        # The server may close the connection while we're still sending
        # This could trigger either BodyError or RequestError depending on timing
        try:
            response = client.post(f"{test_server}/broken-body", content=large_body)
            # If we get here, the request completed - try to read the response
            _ = response.content
        except BodyError as e:
            assert isinstance(e, PrimpError)


class TestUpgradeError:
    """Test UpgradeError exception.

    UpgradeError is raised when there's an error during HTTP protocol upgrade
    (e.g., WebSocket upgrade). This occurs when calling response.upgrade() on a
    response that indicated a protocol upgrade (101 Switching Protocols) but
    the upgrade handshake failed.

    Note: The upgrade() method is not currently exposed in the Python bindings,
    so UpgradeError cannot be reliably triggered with real HTTP requests.
    Therefore, no real tests are available for this exception type.
    """

    # No tests available - upgrade() is not exposed in Python bindings
    # UpgradeError cannot be reliably triggered with real HTTP requests
    pass


class TestAsyncExceptions:
    """Test exceptions with async client."""

    @pytest.mark.asyncio
    async def test_invalid_url_async(self):
        """Invalid URL should raise BuilderError with async client."""
        client = primp.AsyncClient()
        with pytest.raises(BuilderError):
            await client.get("example.com")

    @pytest.mark.asyncio
    async def test_invalid_header_async(self, test_server):
        """Invalid header should raise BuilderError with async client."""
        client = primp.AsyncClient()
        with pytest.raises(BuilderError):
            await client.get(f"{test_server}/get", headers={"X-Test": "value\nwith-newline"})

    @pytest.mark.asyncio
    async def test_connect_error_async(self):
        """Connection error should raise ConnectError with async client."""
        client = primp.AsyncClient()
        with pytest.raises(ConnectError):
            await client.get("http://localhost:1", timeout=5)

    @pytest.mark.asyncio
    async def test_timeout_error_async(self, test_server):
        """Timeout should raise TimeoutError with async client."""
        client = primp.AsyncClient()
        with pytest.raises(TimeoutError):
            await client.get(f"{test_server}/delay/5", timeout=0.5)

    @pytest.mark.asyncio
    async def test_status_error_async(self, test_server):
        """HTTP error status should raise StatusError with async client."""
        client = primp.AsyncClient()
        response = await client.get(f"{test_server}/status/404")
        with pytest.raises(StatusError):
            response.raise_for_status()

    @pytest.mark.asyncio
    async def test_status_error_async_500(self, test_server):
        """HTTP 500 should raise StatusError with async client."""
        client = primp.AsyncClient()
        response = await client.get(f"{test_server}/status/500")
        with pytest.raises(StatusError) as exc_info:
            response.raise_for_status()
        assert exc_info.value.args[0] == 500

    @pytest.mark.asyncio
    async def test_redirect_error_async(self, test_server):
        """Exceeding max_redirects should raise RedirectError with async client."""
        client = primp.AsyncClient(follow_redirects=True, max_redirects=2)
        with pytest.raises(RedirectError):
            await client.get(f"{test_server}/redirect/10")

    @pytest.mark.asyncio
    async def test_decode_error_async(self, test_server):
        """Invalid gzip content should raise DecodeError with async client."""
        client = primp.AsyncClient()
        response = await client.get(f"{test_server}/invalid-gzip")
        with pytest.raises(DecodeError):
            _ = response.content

    @pytest.mark.asyncio
    async def test_can_catch_as_primp_error_async(self):
        """Exceptions should be catchable as PrimpError with async client."""
        client = primp.AsyncClient()
        with pytest.raises(PrimpError):
            await client.get("example.com")

    @pytest.mark.asyncio
    async def test_body_error_async(self, test_server):
        """Sending a large request body to a server that closes early may trigger BodyError with async client.

        This test attempts to trigger BodyError by sending a large request body to a server
        that closes the connection immediately. Depending on timing, this may result in
        BodyError, RequestError, or no error if the server responds before the body is sent.
        """
        client = primp.AsyncClient(timeout=2)
        # Create a large body that will take time to send
        large_body = b"x" * 10_000_000  # 10MB of data
        # The server may close the connection while we're still sending
        try:
            response = await client.post(f"{test_server}/broken-body", content=large_body)
            _ = response.content
        except (BodyError, RequestError, ConnectError) as e:
            # Any of these errors is acceptable - the server closed the connection
            assert isinstance(e, PrimpError)



"""
Type stubs for primp - HTTP client that can impersonate web browsers.

This module provides both synchronous and asynchronous HTTP clients with
browser impersonation capabilities.
"""

from collections.abc import Mapping
from typing import (
    Any,
)

# =============================================================================
# Type Aliases
# =============================================================================

HeadersType = Mapping[str, str] | None
ParamsType = Mapping[str, str] | None
CookiesType = Mapping[str, str] | None
FilesType = Mapping[str, str] | None
AuthType = tuple[str, str | None] | None

# =============================================================================
# Exception Classes
# =============================================================================

class PrimpError(Exception):
    """Base exception for all primp errors."""

    ...

class BuilderError(PrimpError):
    """Error during client builder configuration."""

    ...

class RequestError(PrimpError):
    """Error during request execution."""

    ...

class ConnectError(RequestError):
    """Error connecting to the server."""

    ...

class TimeoutError(RequestError):
    """Request timed out."""

    ...

class StatusError(PrimpError):
    """HTTP status error (4xx/5xx)."""

    ...

class RedirectError(PrimpError):
    """Error during redirect handling."""

    ...

class BodyError(PrimpError):
    """Error with request/response body."""

    ...

class DecodeError(PrimpError):
    """Error decoding response content."""

    ...

class UpgradeError(PrimpError):
    """Error during protocol upgrade."""

    ...

# =============================================================================
# Response Classes (Synchronous)
# =============================================================================

class Response:
    """
    HTTP response object.

    Provides access to response status, headers, cookies, and body content.
    """

    # Properties (read-only)
    url: str
    status_code: int

    @property
    def content(self) -> bytes: ...
    @property
    def encoding(self) -> str: ...
    @encoding.setter
    def encoding(self, value: str | None) -> None: ...
    @property
    def headers(self) -> dict[str, str]: ...
    @property
    def cookies(self) -> dict[str, str]: ...
    @property
    def text(self) -> str: ...
    @property
    def text_markdown(self) -> str: ...
    @property
    def text_plain(self) -> str: ...
    @property
    def text_rich(self) -> str: ...

    # Methods
    def json(self) -> Any: ...
    def raise_for_status(self) -> None: ...

    # Streaming methods
    def read(self) -> bytes: ...
    def iter_bytes(self, chunk_size: int | None = None) -> BytesIterator: ...
    def iter_text(self, chunk_size: int | None = None) -> TextIterator: ...
    def iter_lines(self) -> LinesIterator: ...
    def next(self) -> bytes | None: ...
    def close(self) -> None: ...

    # Context manager
    def __enter__(self) -> Response: ...
    def __exit__(
        self,
        exc_type: Any | None,
        exc_value: Any | None,
        traceback: Any | None,
    ) -> bool: ...

class BytesIterator:
    """Iterator over byte chunks from a streaming response."""

    def __iter__(self) -> BytesIterator: ...
    def __next__(self) -> bytes: ...

class TextIterator:
    """Iterator over text chunks from a streaming response."""

    def __iter__(self) -> TextIterator: ...
    def __next__(self) -> str: ...

class LinesIterator:
    """Iterator over lines from a streaming response."""

    def __iter__(self) -> LinesIterator: ...
    def __next__(self) -> str: ...

# =============================================================================
# Response Classes (Asynchronous)
# =============================================================================

class AsyncResponse:
    """
    Asynchronous HTTP response object.

    Provides access to response status, headers, cookies, and body content.
    All property access is synchronous but may block.
    """

    # Properties (read-only)
    url: str
    status_code: int

    @property
    def content(self) -> bytes: ...
    @property
    def encoding(self) -> str: ...
    @encoding.setter
    def encoding(self, value: str | None) -> None: ...
    @property
    def headers(self) -> dict[str, str]: ...
    @property
    def cookies(self) -> dict[str, str]: ...
    @property
    def text(self) -> str: ...
    @property
    def text_markdown(self) -> str: ...
    @property
    def text_plain(self) -> str: ...
    @property
    def text_rich(self) -> str: ...

    # Methods
    def json(self) -> Any: ...
    def raise_for_status(self) -> None: ...

    # Async streaming methods
    async def aread(self) -> bytes: ...
    def aiter_bytes(self, chunk_size: int | None = None) -> AsyncBytesIterator: ...
    def aiter_text(self, chunk_size: int | None = None) -> AsyncTextIterator: ...
    def aiter_lines(self) -> AsyncLinesIterator: ...
    async def anext(self) -> bytes | None: ...
    async def aclose(self) -> None: ...

    # Async context manager
    async def __aenter__(self) -> AsyncResponse: ...
    async def __aexit__(
        self,
        exc_type: Any | None,
        exc_value: Any | None,
        traceback: Any | None,
    ) -> bool: ...

class AsyncBytesIterator:
    """Async iterator over byte chunks from a streaming response."""

    def __aiter__(self) -> AsyncBytesIterator: ...
    async def __anext__(self) -> bytes: ...

class AsyncTextIterator:
    """Async iterator over text chunks from a streaming response."""

    def __aiter__(self) -> AsyncTextIterator: ...
    async def __anext__(self) -> str: ...

class AsyncLinesIterator:
    """Async iterator over lines from a streaming response."""

    def __aiter__(self) -> AsyncLinesIterator: ...
    async def __anext__(self) -> str: ...

# =============================================================================
# Client Classes
# =============================================================================

class Client:
    """
    Synchronous HTTP client that can impersonate web browsers.

    This client supports browser impersonation, proxy configuration,
    cookie management, and various authentication methods.

    Example:
        >>> client = Client(impersonate="chrome_146")
        >>> response = client.get("https://example.com")
        >>> print(response.text)
    """

    # Instance properties
    auth: tuple[str, str | None] | None
    auth_bearer: str | None
    params: Mapping[str, str] | None
    proxy: str | None
    timeout: float | None
    connect_timeout: float | None
    read_timeout: float | None
    base_url: str | None

    @property
    def headers(self) -> dict[str, str]: ...
    @headers.setter
    def headers(self, value: Mapping[str, str] | None) -> None: ...
    @property
    def impersonate(self) -> str | None: ...
    @property
    def impersonate_os(self) -> str | None: ...
    def __init__(
        self,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookie_store: bool = True,
        referer: bool = True,
        proxy: str | None = None,
        timeout: float | None = None,
        connect_timeout: float | None = None,
        read_timeout: float | None = None,
        impersonate: str | None = None,
        impersonate_os: str | None = None,
        follow_redirects: bool = True,
        max_redirects: int = 20,
        verify: bool = True,
        ca_cert_file: str | None = None,
        https_only: bool = False,
        http2_only: bool = False,
        base_url: str | None = None,
        cookies: Mapping[str, str] | None = None,
    ) -> None: ...
    def headers_update(self, new_headers: Mapping[str, str] | None) -> None: ...
    def get_cookies(self, url: str) -> dict[str, str]: ...
    def set_cookies(self, url: str, cookies: Mapping[str, str] | None) -> None: ...
    def request(
        self,
        method: str,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> Response: ...
    def get(
        self,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> Response: ...
    def head(
        self,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> Response: ...
    def options(
        self,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> Response: ...
    def delete(
        self,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> Response: ...
    def post(
        self,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> Response: ...
    def put(
        self,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> Response: ...
    def patch(
        self,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> Response: ...

    # Context manager
    def __enter__(self) -> Client: ...
    def __exit__(
        self,
        exc_type: Any | None,
        exc_value: Any | None,
        traceback: Any | None,
    ) -> None: ...

class AsyncClient:
    """
    Asynchronous HTTP client that can impersonate web browsers.

    This client supports browser impersonation, proxy configuration,
    cookie management, and various authentication methods.

    Example:
        >>> async with AsyncClient(impersonate="chrome_146") as client:
        ...     response = await client.get("https://example.com")
        ...     print(response.text)
    """

    # Instance properties
    auth: tuple[str, str | None] | None
    auth_bearer: str | None
    params: Mapping[str, str] | None
    proxy: str | None
    timeout: float | None
    connect_timeout: float | None
    read_timeout: float | None
    base_url: str | None

    @property
    def headers(self) -> dict[str, str]: ...
    @headers.setter
    def headers(self, value: Mapping[str, str] | None) -> None: ...
    @property
    def impersonate(self) -> str | None: ...
    @property
    def impersonate_os(self) -> str | None: ...
    def __init__(
        self,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookie_store: bool = True,
        referer: bool = True,
        proxy: str | None = None,
        timeout: float | None = None,
        connect_timeout: float | None = None,
        read_timeout: float | None = None,
        impersonate: str | None = None,
        impersonate_os: str | None = None,
        follow_redirects: bool = True,
        max_redirects: int = 20,
        verify: bool = True,
        ca_cert_file: str | None = None,
        https_only: bool = False,
        http2_only: bool = False,
        base_url: str | None = None,
        cookies: Mapping[str, str] | None = None,
    ) -> None: ...
    def headers_update(self, new_headers: Mapping[str, str] | None) -> None: ...
    def get_cookies(self, url: str) -> dict[str, str]: ...
    def set_cookies(self, url: str, cookies: Mapping[str, str] | None) -> None: ...
    async def request(
        self,
        method: str,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> AsyncResponse: ...
    async def get(
        self,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> AsyncResponse: ...
    async def head(
        self,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> AsyncResponse: ...
    async def options(
        self,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> AsyncResponse: ...
    async def delete(
        self,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> AsyncResponse: ...
    async def post(
        self,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> AsyncResponse: ...
    async def put(
        self,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> AsyncResponse: ...
    async def patch(
        self,
        url: str,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | None = None,
        content: bytes | None = None,
        data: Any | None = None,
        json: Any | None = None,
        files: Mapping[str, str] | None = None,
        auth: tuple[str, str | None] | None = None,
        auth_bearer: str | None = None,
        timeout: float | None = None,
        read_timeout: float | None = None,
        follow_redirects: bool | None = None,
        stream: bool = False,
    ) -> AsyncResponse: ...

    # Async context manager
    async def __aenter__(self) -> AsyncClient: ...
    async def __aexit__(
        self,
        exc_type: Any | None,
        exc_value: Any | None,
        traceback: Any | None,
    ) -> None: ...

# =============================================================================
# Module-level Functions
# =============================================================================

def get(
    url: str,
    params: Mapping[str, str] | None = None,
    headers: Mapping[str, str] | None = None,
    cookies: Mapping[str, str] | None = None,
    content: bytes | None = None,
    data: Any | None = None,
    json: Any | None = None,
    files: Mapping[str, str] | None = None,
    auth: tuple[str, str | None] | None = None,
    auth_bearer: str | None = None,
    timeout: float | None = None,
    connect_timeout: float | None = None,
    read_timeout: float | None = None,
    impersonate: str | None = None,
    impersonate_os: str | None = None,
    verify: bool = True,
    ca_cert_file: str | None = None,
    follow_redirects: bool | None = None,
    stream: bool = False,
) -> Response:
    """Send a GET request with a temporary client."""
    ...

def head(
    url: str,
    params: Mapping[str, str] | None = None,
    headers: Mapping[str, str] | None = None,
    cookies: Mapping[str, str] | None = None,
    content: bytes | None = None,
    data: Any | None = None,
    json: Any | None = None,
    files: Mapping[str, str] | None = None,
    auth: tuple[str, str | None] | None = None,
    auth_bearer: str | None = None,
    timeout: float | None = None,
    connect_timeout: float | None = None,
    read_timeout: float | None = None,
    impersonate: str | None = None,
    impersonate_os: str | None = None,
    verify: bool = True,
    ca_cert_file: str | None = None,
    follow_redirects: bool | None = None,
    stream: bool = False,
) -> Response:
    """Send a HEAD request with a temporary client."""
    ...

def options(
    url: str,
    params: Mapping[str, str] | None = None,
    headers: Mapping[str, str] | None = None,
    cookies: Mapping[str, str] | None = None,
    content: bytes | None = None,
    data: Any | None = None,
    json: Any | None = None,
    files: Mapping[str, str] | None = None,
    auth: tuple[str, str | None] | None = None,
    auth_bearer: str | None = None,
    timeout: float | None = None,
    connect_timeout: float | None = None,
    read_timeout: float | None = None,
    impersonate: str | None = None,
    impersonate_os: str | None = None,
    verify: bool = True,
    ca_cert_file: str | None = None,
    follow_redirects: bool | None = None,
    stream: bool = False,
) -> Response:
    """Send an OPTIONS request with a temporary client."""
    ...

def delete(
    url: str,
    params: Mapping[str, str] | None = None,
    headers: Mapping[str, str] | None = None,
    cookies: Mapping[str, str] | None = None,
    content: bytes | None = None,
    data: Any | None = None,
    json: Any | None = None,
    files: Mapping[str, str] | None = None,
    auth: tuple[str, str | None] | None = None,
    auth_bearer: str | None = None,
    timeout: float | None = None,
    connect_timeout: float | None = None,
    read_timeout: float | None = None,
    impersonate: str | None = None,
    impersonate_os: str | None = None,
    verify: bool = True,
    ca_cert_file: str | None = None,
    follow_redirects: bool | None = None,
    stream: bool = False,
) -> Response:
    """Send a DELETE request with a temporary client."""
    ...

def post(
    url: str,
    params: Mapping[str, str] | None = None,
    headers: Mapping[str, str] | None = None,
    cookies: Mapping[str, str] | None = None,
    content: bytes | None = None,
    data: Any | None = None,
    json: Any | None = None,
    files: Mapping[str, str] | None = None,
    auth: tuple[str, str | None] | None = None,
    auth_bearer: str | None = None,
    timeout: float | None = None,
    connect_timeout: float | None = None,
    read_timeout: float | None = None,
    impersonate: str | None = None,
    impersonate_os: str | None = None,
    verify: bool = True,
    ca_cert_file: str | None = None,
    follow_redirects: bool | None = None,
    stream: bool = False,
) -> Response:
    """Send a POST request with a temporary client."""
    ...

def put(
    url: str,
    params: Mapping[str, str] | None = None,
    headers: Mapping[str, str] | None = None,
    cookies: Mapping[str, str] | None = None,
    content: bytes | None = None,
    data: Any | None = None,
    json: Any | None = None,
    files: Mapping[str, str] | None = None,
    auth: tuple[str, str | None] | None = None,
    auth_bearer: str | None = None,
    timeout: float | None = None,
    connect_timeout: float | None = None,
    read_timeout: float | None = None,
    impersonate: str | None = None,
    impersonate_os: str | None = None,
    verify: bool = True,
    ca_cert_file: str | None = None,
    follow_redirects: bool | None = None,
    stream: bool = False,
) -> Response:
    """Send a PUT request with a temporary client."""
    ...

def patch(
    url: str,
    params: Mapping[str, str] | None = None,
    headers: Mapping[str, str] | None = None,
    cookies: Mapping[str, str] | None = None,
    content: bytes | None = None,
    data: Any | None = None,
    json: Any | None = None,
    files: Mapping[str, str] | None = None,
    auth: tuple[str, str | None] | None = None,
    auth_bearer: str | None = None,
    timeout: float | None = None,
    connect_timeout: float | None = None,
    read_timeout: float | None = None,
    impersonate: str | None = None,
    impersonate_os: str | None = None,
    verify: bool = True,
    ca_cert_file: str | None = None,
    follow_redirects: bool | None = None,
    stream: bool = False,
) -> Response:
    """Send a PATCH request with a temporary client."""
    ...

def request(
    method: str,
    url: str,
    params: Mapping[str, str] | None = None,
    headers: Mapping[str, str] | None = None,
    cookies: Mapping[str, str] | None = None,
    content: bytes | None = None,
    data: Any | None = None,
    json: Any | None = None,
    files: Mapping[str, str] | None = None,
    auth: tuple[str, str | None] | None = None,
    auth_bearer: str | None = None,
    timeout: float | None = None,
    connect_timeout: float | None = None,
    read_timeout: float | None = None,
    impersonate: str | None = None,
    impersonate_os: str | None = None,
    verify: bool = True,
    ca_cert_file: str | None = None,
    follow_redirects: bool | None = None,
    stream: bool = False,
) -> Response:
    """Send a request with a custom HTTP method using a temporary client."""
    ...



"""
Tests for Response properties and methods.

This module tests:
- url: Response URL
- status_code: HTTP status code
- content: Raw bytes content
- encoding: Response encoding
- text: Text content
- headers: Response headers
- cookies: Response cookies
- text_markdown: Markdown conversion
- text_plain: Plain text extraction
- text_rich: Rich text conversion
- json(): JSON parsing
- raise_for_status(): Status code handling
"""

import pytest

import primp


class TestResponseUrl:
    """Tests for Response.url property."""
    
    def test_sync_client_response_url(self, test_server: str) -> None:
        """Test response URL property on sync Client."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/get")
        
        assert response.url is not None
        assert "/get" in response.url
    
    @pytest.mark.asyncio
    async def test_async_client_response_url(self, test_server: str) -> None:
        """Test response URL property on AsyncClient."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/get")
        
        assert response.url is not None
        assert "/get" in response.url


class TestResponseStatusCode:
    """Tests for Response.status_code property."""
    
    def test_sync_client_response_status_code(self, test_server: str) -> None:
        """Test response status_code property on sync Client."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/get")
        
        assert response.status_code == 200
        assert isinstance(response.status_code, int)
    
    @pytest.mark.asyncio
    async def test_async_client_response_status_code(self, test_server: str) -> None:
        """Test response status_code property on AsyncClient."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/get")
        
        assert response.status_code == 200
        assert isinstance(response.status_code, int)
    
    def test_sync_client_response_status_code_404(self, test_server: str) -> None:
        """Test response status_code for 404 on sync Client."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/status/404")
        
        assert response.status_code == 404
    
    @pytest.mark.asyncio
    async def test_async_client_response_status_code_404(self, test_server: str) -> None:
        """Test response status_code for 404 on AsyncClient."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/status/404")
        
        assert response.status_code == 404


class TestResponseContent:
    """Tests for Response.content property."""
    
    def test_sync_client_response_content(self, test_server: str) -> None:
        """Test response content property on sync Client."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/get")
        
        assert response.content is not None
        assert isinstance(response.content, bytes)
        assert len(response.content) > 0
    
    @pytest.mark.asyncio
    async def test_async_client_response_content(self, test_server: str) -> None:
        """Test response content property on AsyncClient."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/get")
        
        assert response.content is not None
        assert isinstance(response.content, bytes)
        assert len(response.content) > 0


class TestResponseEncoding:
    """Tests for Response.encoding property."""
    
    def test_sync_client_response_encoding(self, test_server: str) -> None:
        """Test response encoding property on sync Client."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/get")
        
        # Encoding may be None or a string
        if response.encoding is not None:
            assert isinstance(response.encoding, str)
    
    @pytest.mark.asyncio
    async def test_async_client_response_encoding(self, test_server: str) -> None:
        """Test response encoding property on AsyncClient."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/get")
        
        if response.encoding is not None:
            assert isinstance(response.encoding, str)
    
    def test_sync_client_response_encoding_setter(self, test_server: str) -> None:
        """Test response encoding property setter on sync Client."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/get")
        
        response.encoding = "utf-8"
        assert response.encoding == "utf-8"
    
    @pytest.mark.asyncio
    async def test_async_client_response_encoding_setter(self, test_server: str) -> None:
        """Test response encoding property setter on AsyncClient."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/get")
        
        response.encoding = "utf-8"
        assert response.encoding == "utf-8"


class TestResponseText:
    """Tests for Response.text property."""
    
    def test_sync_client_response_text(self, test_server: str) -> None:
        """Test response text property on sync Client."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/get")
        
        assert response.text is not None
        assert isinstance(response.text, str)
        assert len(response.text) > 0
    
    @pytest.mark.asyncio
    async def test_async_client_response_text(self, test_server: str) -> None:
        """Test response text property on AsyncClient."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/get")
        
        assert response.text is not None
        assert isinstance(response.text, str)
        assert len(response.text) > 0
    
    def test_sync_client_response_text_html(self, test_server: str) -> None:
        """Test response text property for HTML content."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/html")
        
        assert response.status_code == 200
        text = response.text
        assert "Welcome to Test Server" in text or "Test Page" in text


class TestResponseHeaders:
    """Tests for Response.headers property."""
    
    def test_sync_client_response_headers(self, test_server: str) -> None:
        """Test response headers property on sync Client."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/get")
        
        assert response.headers is not None
        assert isinstance(response.headers, dict)
        # Check for common headers
        assert "content-type" in response.headers or "Content-Type" in response.headers
    
    @pytest.mark.asyncio
    async def test_async_client_response_headers(self, test_server: str) -> None:
        """Test response headers property on AsyncClient."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/get")
        
        assert response.headers is not None
        assert isinstance(response.headers, dict)


class TestResponseCookies:
    """Tests for Response.cookies property."""
    
    def test_sync_client_response_cookies(self, test_server: str) -> None:
        """Test response cookies property on sync Client."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/cookies/set?test_cookie=test_value")
        
        # Response should have cookies
        cookies = response.cookies
        assert cookies is not None
    
    @pytest.mark.asyncio
    async def test_async_client_response_cookies(self, test_server: str) -> None:
        """Test response cookies property on AsyncClient."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/cookies/set?test_cookie=test_value")
        
        cookies = response.cookies
        assert cookies is not None


class TestResponseTextMarkdown:
    """Tests for Response.text_markdown property."""
    
    def test_sync_client_response_text_markdown(self, test_server: str) -> None:
        """Test response text_markdown property on sync Client."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/html")
        
        assert response.status_code == 200
        # text_markdown should convert HTML to markdown
        markdown = response.text_markdown
        assert markdown is not None
        assert isinstance(markdown, str)
    
    @pytest.mark.asyncio
    async def test_async_client_response_text_markdown(self, test_server: str) -> None:
        """Test response text_markdown property on AsyncClient."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/html")
        
        assert response.status_code == 200
        markdown = response.text_markdown
        assert markdown is not None
        assert isinstance(markdown, str)


class TestResponseTextPlain:
    """Tests for Response.text_plain property."""
    
    def test_sync_client_response_text_plain(self, test_server: str) -> None:
        """Test response text_plain property on sync Client."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/html")
        
        assert response.status_code == 200
        # text_plain should extract plain text from HTML
        plain = response.text_plain
        assert plain is not None
        assert isinstance(plain, str)
    
    @pytest.mark.asyncio
    async def test_async_client_response_text_plain(self, test_server: str) -> None:
        """Test response text_plain property on AsyncClient."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/html")
        
        assert response.status_code == 200
        plain = response.text_plain
        assert plain is not None
        assert isinstance(plain, str)


class TestResponseTextRich:
    """Tests for Response.text_rich property."""
    
    def test_sync_client_response_text_rich(self, test_server: str) -> None:
        """Test response text_rich property on sync Client."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/html")
        
        assert response.status_code == 200
        # text_rich should convert HTML to rich text
        rich = response.text_rich
        assert rich is not None
    
    @pytest.mark.asyncio
    async def test_async_client_response_text_rich(self, test_server: str) -> None:
        """Test response text_rich property on AsyncClient."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/html")
        
        assert response.status_code == 200
        rich = response.text_rich
        assert rich is not None


class TestResponseJson:
    """Tests for Response.json() method."""
    
    def test_sync_client_response_json(self, test_server: str) -> None:
        """Test response json() method on sync Client."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/get")
        
        assert response.status_code == 200
        data = response.json()
        
        assert isinstance(data, dict)
        assert "method" in data
        assert data["method"] == "GET"
    
    @pytest.mark.asyncio
    async def test_async_client_response_json(self, test_server: str) -> None:
        """Test response json() method on AsyncClient."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/get")
        
        assert response.status_code == 200
        data = response.json()
        
        assert isinstance(data, dict)
        assert "method" in data
        assert data["method"] == "GET"
    
    def test_sync_client_response_json_nested(self, test_server: str) -> None:
        """Test response json() method with nested data."""
        base_url = test_server
        
        client = primp.Client()
        json_data = {
            "level1": {
                "level2": {
                    "level3": "deep_value"
                }
            },
            "array": [1, 2, 3]
        }
        response = client.post(f"{base_url}/post", json=json_data)
        
        assert response.status_code == 200
        data = response.json()
        
        assert data["json"]["level1"]["level2"]["level3"] == "deep_value"
        assert data["json"]["array"] == [1, 2, 3]


class TestResponseRaiseForStatus:
    """Tests for Response.raise_for_status() method."""
    
    def test_sync_client_raise_for_status_success(self, test_server: str) -> None:
        """Test raise_for_status() does not raise for successful status."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/get")
        
        # Should not raise
        response.raise_for_status()
    
    @pytest.mark.asyncio
    async def test_async_client_raise_for_status_success(self, test_server: str) -> None:
        """Test raise_for_status() does not raise for successful status."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/get")
        
        # Should not raise
        response.raise_for_status()
    
    def test_sync_client_raise_for_status_404(self, test_server: str) -> None:
        """Test raise_for_status() raises for 404 status."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/status/404")
        
        assert response.status_code == 404
        
        with pytest.raises(Exception):
            response.raise_for_status()
    
    @pytest.mark.asyncio
    async def test_async_client_raise_for_status_404(self, test_server: str) -> None:
        """Test raise_for_status() raises for 404 status."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/status/404")
        
        assert response.status_code == 404
        
        with pytest.raises(Exception):
            response.raise_for_status()
    
    def test_sync_client_raise_for_status_500(self, test_server: str) -> None:
        """Test raise_for_status() raises for 500 status."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/status/500")
        
        assert response.status_code == 500
        
        with pytest.raises(Exception):
            response.raise_for_status()
    
    @pytest.mark.asyncio
    async def test_async_client_raise_for_status_500(self, test_server: str) -> None:
        """Test raise_for_status() raises for 500 status."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/status/500")
        
        assert response.status_code == 500
        
        with pytest.raises(Exception):
            response.raise_for_status()


class TestResponseStatusCodeRanges:
    """Tests for various HTTP status code ranges."""
    
    @pytest.mark.parametrize("status_code", [200, 201, 202, 204, 301, 302, 400, 401, 403, 404, 500, 502, 503])
    def test_sync_client_status_codes(self, test_server: str, status_code: int) -> None:
        """Test various status codes on sync Client."""
        base_url = test_server
        
        client = primp.Client()
        response = client.get(f"{base_url}/status/{status_code}")
        
        assert response.status_code == status_code
    
    @pytest.mark.parametrize("status_code", [200, 201, 202, 204, 301, 302, 400, 401, 403, 404, 500, 502, 503])
    @pytest.mark.asyncio
    async def test_async_client_status_codes(self, test_server: str, status_code: int) -> None:
        """Test various status codes on AsyncClient."""
        base_url = test_server
        
        client = primp.AsyncClient()
        response = await client.get(f"{base_url}/status/{status_code}")
        
        assert response.status_code == status_code
